// ------------------------------------------------
// SIMULATION OF A GALAXY
// Using Nearest Grid Point / Particle-Mesh method
// ------------------------------------------------
//
// galaxy.cpp
//
// Written by Andrew Wedgbury
// Copyright 1999
// All Rights Reserved
//
#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#pragma hdrstop

//
// WINDOWS INTERFACE DATA
//
HINSTANCE hInst; // INSTANCE HANDLE
HWND hDispWin,hControlWin; // WINDOW HANDLES
HANDLE hCalcThread; // CALCULATION THREAD HANDLE

//
// SIMULATION DATA TYPES
//
struct VECTOR {
  VECTOR(double a=0, double b=0) : x(a), y(b) { };
  double x; // X-COMPONENT
  double y; // Y-COMPONENT
};

struct MESHCELL {
  MESHCELL() {};
  VECTOR g; // GRAVITATIONAL FIELD IN X AND Y [L/TT]
  double m; // TOTAL MASS IN CELL [M]
};

struct STAR {
  STAR() {};
  VECTOR r; // POSITION OF STAR [L]
  VECTOR V; // VELOCITY OF STAR [L/T]
  double m; // MASS OF STAR [M]
  int c;    // TYPE OF STAR
};

struct PIXEL {
  PIXEL() : x(0), y(0) {};
  WORD x,y;
};

//
// PHYSICAL CONSTANTS AND UNITS
//
#define M (double)1.99e30                     // Mass unit (solar mass)
#define L (double)3.08e19                     // Length unit (Kiloparsec)
#define T (double)3.15569259747e7             // Time unit (Year)
const double G = 6.67259e-11 * T*T*M/(L*L*L); // Gravitational constant [LLL/TTM]
#define PI 3.14159265359                      // PI
const double TWOPI =2.0*PI;                   // PRECALCULATED 2*PI

//
// PARTICLE DATA
//
#define NMAX 250000                           // MAXIMUM NUMBER OF STARS
int N = 100000;                               // NUMBER OF STARS IN SIMULATION
double D = 1.0e6;                             // TIMESTEP [T]
DWORD t = 0;                                  // TIME [steps]
double R0 = 15.0;                             // INITIAL DISK RADIUS [L]
double Rc = 5.0;                             // INITIAL CENTRAL HALO RADIUS [L]
double MGX = 1e11;                            // MASS oF GALAXY [M]
double M0 = MGX/N;                            // STAR MASS [M]
double Mh = 0e12;                             // CENTRAL HALO MASS [M]
double VDISP = 0.0e-8;                        // VELOCITY DISPERSION [L/T]
STAR stars[NMAX];                             // STARS ARRAY
//int nDiscType = 5;                          // INITIAL DISC TYPE (see SetInitial)
int nDiscType = 5;                            // INITIAL DISC TYPE (see SetInitial)
int nHaloType = 2;                            // HALO TYPE (see CalcHaloField)

//
// MESH DATA
//
#define CMAX 256                              // MAXIMUM EDGE DIMENSION (CELLS)
int C = 100;                                  // MESH EDGE DIMENSION (CELLS)
double S = 60.0;                              // MESH SIZE [L]
double d = S/C;                               // MESH SPACING [L]
double CM = (d*C/2);                          // CENTER OF MESH [L]
MESHCELL mesh[CMAX][CMAX];                    // MESH ARRAY
double RxCache[CMAX][CMAX];                   // ARRAY FOR CACHED Rx VALUES
double RyCache[CMAX][CMAX];                   // ARRAY FOR CACHED Ry VALUES
PIXEL meshused[CMAX*CMAX];                    // STORES COORDINATES OF MESH CELLS BEING USED

//
// FILE OUTPUT DATA
//
int nEdgeCollisions=0;                        // COUNTER FOR EDGE COLLISIONS
int nCellsUsed=0;                             // COUNTER FOR NUMBER OF CELLS OCCUPIED
char szBuffer[512];                           // TEXT BUFFER FOR O/P
char szDataFile[MAX_PATH];                    // DATA O/P FILE NAME
double OPINTDATA = 0.0e7;                     // O/P INTERVAL FOR CONFIG DATA [T] (0 to disable)
int hDataFile = NULL;                         // DATA O/P FILE HANDLE
int hGraphFile = NULL;
double OPINTBIN = 1.0e7;                      // O/P INTERVAL FOR BINARY DATA [T] (0 to disable)
int hBinFile = NULL;                          // BINARY O/P FILE HANDLE
PIXEL binout[NMAX];                           // BINARY O/P BUFFER

//
// GRAPHICS OUTPUT DATA
//
int RESX = 1024;                              // SCREEN X RESOLUTION [pixels]
int RESY = 768;                               // SCREEN Y RESOLUTION [pixels]
int WINDOWX = RESY-20;                        // O/P WINDOW X DIMENSION [pixels]
int WINDOWY = RESY-20;                        // O/P WINDOW Y DIMENSION [pixels]
int SIDEX = RESY;                             // X POSITION OF SIDE INFO BAR [pixels]
double xs = WINDOWX/S;                        // X SCALE [pixels/L]
double ys = WINDOWY/S;                        // Y SCALE [pixels/L]
int xt = 0;                                   // X TRANSFORM
int yt = 0;                                   // Y TRANSFORM
double vmax = 0;                              // SCALE FOR VELOCITY GRAPH
double dmax = 0;                              // SCALE FOR DENSITY GRAPH
double dbins[400];                            // RADIAL DENSITY ARRAY
double fmax = 0;                              // SCALE FOR FIELD COLOURING
#define NIS 2                                 // NUMBER OF INDICATOR STARS
VECTOR Ir[NIS];                               // INDICATOR STAR POSITIONS
double nRevs[NIS];                            // NUMBER OF REVOLUTIONS OF INDICATORS
double Ep = 0.0;                              // TOTAL POTENTIAL ENERGY ACCUMULATOR
double Ek = 0.0;                              // TOTAL KINETIC ENERGY ACCUMULATOR
DWORD dwS=0,dwF=0;

//
// SIMULATION CONTROL FLAGS
//
BOOL bDrawField = 0;                          // DRAW FIELD FLAG
BOOL bDrawStars = 1;                          // DRAW STARS FLAG
BOOL bDrawText = 1;                           // DRAW TEXT PANEL FLAG
BOOL bDrawIndic = 1;                          // DRAW INDICATOR STARS FLAG
BOOL bDrawAxes = 1;                           // DRAW AXES FLAG
BOOL bDrawHalo = 0;                           // DRAW HALO FLAG
BOOL bPaused = 0;                             // PAUSE / GO FLAG

//
// FUNCTION PROTOTYPES
//
void Timestep(void);
void CalcMesh(void);
void CheckBounds(int);
void OutputData(void);
void WriteDataFile(void);
void SetInitial(void);
VECTOR CalcMeshField(double x, double y);
void AddStar(double x, double y, double m);
VECTOR CalcHaloField(double,double);
double CalcEnergy(int);
void InitRCache(void);
double LookupRx(int,int);
double LookupRy(int,int);
double radius(double,double);

//
// FUNCTION: Timestep
// Performs one timestep of the simulation
//
void Timestep(void)
{
  if (t==0) {
    SetInitial();
    OutputData();

  } else {
    CalcMesh();

  }
  
  // Increment timestep counter
  t++;
  printf("%u ",t);

  // Loop over all stars
  for (int is=0; is<N; is++) {

    // Check star is within mesh
    CheckBounds(is);

    // Save previous velocities
    double pvx = stars[is].V.x;
    double pvy = stars[is].V.y;
    
    VECTOR gh = CalcHaloField(stars[is].r.x,stars[is].r.y);
    VECTOR gm = CalcMeshField(stars[is].r.x,stars[is].r.y);

    // Update particle velocities from halo and mesh
    stars[is].V.x += (gm.x + gh.x) * D;
    stars[is].V.y += (gm.y + gh.y) * D;

    // Update particle positions
    stars[is].r.x += (pvx + stars[is].V.x) * 0.5 *D;
    stars[is].r.y += (pvy + stars[is].V.y) * 0.5 *D;
  }

  OutputData();
}

//
// FUNCTION: random
// Returns a random integer between 0 and max
// (note: has a resolution of 32768)
//
inline int random(int max)
{
  return MulDiv(rand(),max,32768);
}

//
// FUNCTION drandom
// Returns a random double between 0 and max
//
inline double drandom(double max)
{
  return max * rand()/32768.0;
}

//
// FUNCTION gausdev
// Returns a gaussian deviate using Box Muller algorithm
//
double gausdev(void)
{
  double rsq,x,y;
  do {
    x = drandom(2.0)-1.0;
    y = drandom(2.0)-1.0;
    rsq = x*x + y*y;
  }
  while ((rsq >= 1.0) || (rsq == 0.0)) ;
  return x*sqrt(-2.0*log(rsq)/rsq);
}

//
// FUNCTION: radius
// Converts x,y to r using pythagoras' theorem
//
inline double radius(double x, double y)
{
  return sqrt((x*x)+(y*y));
}

//
// INITIAL CONDITIONS TEMPOARY DATA
// (Too big for local storage)
//
double r[NMAX];
double theta[NMAX];

//
// FUNCTION: SetInitial
// Setup initial conditions for stars
//
void SetInitial(void)
{
  int is;
  double u;

  //
  // Assign variables
  //
  M0 = (double)MGX / N;

  //
  // Open files
  //
  if (hDataFile)
    _close(hDataFile);
  if (hBinFile)
    _close(hBinFile);
  if (hGraphFile)
    _close(hGraphFile);
  char szFileName[MAX_PATH];

  if (OPINTDATA > 0.0) {
    lstrcpy(szFileName,szDataFile);
    lstrcat(szFileName,"_t.csv");
    hDataFile = _open(szFileName,_O_WRONLY | _O_CREAT | _O_TRUNC,_S_IREAD | _S_IWRITE);
  }

  if (OPINTBIN > 0.0) {
    lstrcpy(szFileName,szDataFile);
    lstrcat(szFileName,"_g.csv");
    hGraphFile = _open(szFileName,_O_WRONLY | _O_CREAT | _O_TRUNC,_S_IREAD | _S_IWRITE);
  }

  //
  // Set initial star postions
  //
  for (is=0; is<N; is++) {
    //
    // Star mass & type
    //
    stars[is].m = M0;
    stars[is].c = 0;
    //
    // Generate r
    //
    switch (nDiscType) {
      case 1: // UNIFORM SPATIAL DISTRIBUTION
        r[is] = R0*sqrt(drandom(1));
        break;
      case 2: // UNIFORM r
        r[is] = drandom(R0);
        break;
      case 3: // RING
        r[is] = R0;
        break;
      case 4: // GAUSSIAN DENSITY DISTRIBUTION
        do {
          double rand = gausdev();
          r[is] = (R0/4.0)*sqrt(sqrt(rand*rand));
        } while (r[is] > R0);
        break;
      case 5: //  = 0 sqrt(1 - (r/R))
        do {
          u = drandom(1.0);
          r[is] = drandom(1.0);
        } while ( u > sqrt(1.0-(r[is]*r[is])) );
        r[is] = R0*sqrt(r[is]);
      break;
    }
    //
    // Generate random theta between 0 and 2*PI
    //
    theta[is] = drandom(TWOPI);
    //
    // Set star cartesian position from r and theta
    //
    stars[is].r.x = CM + (r[is] * cos(theta[is]));
    stars[is].r.y = CM + (r[is] * sin(theta[is]));
  }

  //
  // Centre star
  //
  theta[0]=0;
  r[0]=0;
  stars[0].r.x = CM;
  stars[0].r.y = CM;
  stars[0].c = 3;

  //
  // Indicator stars
  //
  theta[1]=-PI/2;
  r[1]=Rc;
  stars[1].r.x = Ir[0].x = CM;
  stars[1].r.y = Ir[0].y = CM-r[1];
  stars[1].c = 3;
  theta[2]=-PI/2;
  r[2]=R0;
  stars[2].r.x = Ir[1].x = CM;
  stars[2].r.y = Ir[1].y = CM-r[2];
  stars[2].c = 3;

  //
  // Calculate initial forces to get balancing velocity
  //
  CalcMesh();
  for (is=0; is<N; is++) {

    // Calculate required centripetal velocity and set
    VECTOR gh = CalcHaloField(stars[is].r.x,stars[is].r.y);
    VECTOR gm = CalcMeshField(stars[is].r.x,stars[is].r.y);
    
    double a = radius(gh.x+gm.x, gh.y+gm.y);
    double F = 0.9 * sqrt(a*r[is]);
    
    stars[is].V.x = F * -sin(PI-theta[is]);
    stars[is].V.y = F * -cos(PI-theta[is]);

    //
    // Add gaussian velocity dispersion if required
    //
    if (VDISP) {
      stars[is].V.x += (r[is]/R0)* VDISP * gausdev();
      stars[is].V.y += (r[is]/R0)* VDISP * gausdev();
    }
  }
}

void AddStar(double x, double y, double m)
{
  double icx=0;
  double icy=0;
  double rcx=modf(x/d, &icx);
  double rcy=modf(y/d, &icy);

  int cx = (int)icx;
  int cy = (int)icy;
  if (rcx<0.5) {
    cx--;
    rcx += 0.5;
  } else {
    rcx -= 0.5;
  }
  if (rcy<0.5) {
    cy--;
    rcy += 0.5;
  } else {
    rcy -= 0.5;
  }

  double wa = (1.0-rcx)*(1.0-rcy);
  double wb = rcx*(1.0-rcy);
  double wc = rcx*rcy;
  double wd = (1.0-rcx)*rcy;

  if ( ((cx > 0 ) && (cx < C-1)) && ((cy > 0) && (cy < C-1)) ) {
    mesh[cx][cy].m     += m * wa;
    mesh[cx+1][cy].m   += m * wb;
    mesh[cx+1][cy+1].m += m * wc;
    mesh[cx][cy+1].m   += m * wd;
  }
}

VECTOR CalcMeshField(double x, double y)
{
  double icx=0;
  double icy=0;
  double rcx=modf(x/d, &icx);
  double rcy=modf(y/d, &icy);

  int cx = (int)icx;
  int cy = (int)icy;
  if (rcx<0.5) {
    cx--;
    rcx += 0.5;
  } else {
    rcx -= 0.5;
  }
  if (rcy<0.5) {
    cy--;
    rcy += 0.5;
  } else {
    rcy -= 0.5;
  }

  double wa = (1.0-rcx)*(1.0-rcy);
  double wb = rcx*(1.0-rcy);
  double wc = rcx*rcy;
  double wd = (1.0-rcx)*rcy;

  return VECTOR(
    wa*mesh[cx][cy].g.x + wb*mesh[cx+1][cy].g.x + wc*mesh[cx+1][cy+1].g.x + wd*mesh[cx][cy+1].g.x,
    wa*mesh[cx][cy].g.y + wb*mesh[cx+1][cy].g.y + wc*mesh[cx+1][cy+1].g.y + wd*mesh[cx][cy+1].g.y
  );
}

//
// FUNCTION: CalcHaloField
// Calculate field due to halo
//
VECTOR CalcHaloField(double x,double y)
{
  if (Mh==0.0 || Rc<=0) {
    return VECTOR(0,0);
  }
  
  double R = radius(x-CM,y-CM);
  double f = 0.0;
  double Ru = (x-CM) / R;
  double Rv = (y-CM) / R;
  switch (nHaloType) {
    case 1:
      if (R > Rc) // SOLID SPHERE
        f = -G*Mh / (R*R);
      else
        f = -G*Mh*R / (Rc*Rc*Rc);
      break;
    case 2: // EXPONENTIAL DENSITY SPHERE
      double t1 = Rc*Rc;
      double t4 = R*R;
      double t7 = exp(-R/Rc);
      f = 1/t4*Mh*(t7*t4+2.0*Rc*t7*R+2.0*t1*t7-2.0*t1)/t1*G/2.0;
      break;
  }
  
  return VECTOR(Ru*f, Rv*f);
}

//
// FUNCTION CalcMesh
// Calculate field at mesh points
//
void CalcMesh()
{
  int iu,iv,ix,iy,is;
  nCellsUsed=0;

  //
  // Set all mesh cells to zero mass and field
  //
  for (iu=0; iu<C; iu++) {
    for (iv=0; iv<C; iv++) {
      mesh[iu][iv].m = 0;
      mesh[iu][iv].g.x = 0;
      mesh[iu][iv].g.y = 0;
    }
  }

  //
  // Get mass of stars in each mesh cell
  //
  for (is=0; is<N; is++) {
    
    //
    // Get mesh cell containing star
    //

    AddStar(stars[is].r.x, stars[is].r.y, stars[is].m);
  }

  //
  // Run over mesh and save list of cells containing stars
  // (this saves time and allows bigger meshes)
  //
  for (iu=0; iu<C; iu++) { // Loop over u
    for (iv=0; iv<C; iv++) { // Loop over v
      if (mesh[iu][iv].m || bDrawField) {
        meshused[nCellsUsed].x = iu;
        meshused[nCellsUsed].y = iv;
        nCellsUsed++;
      }
    }
  }

  //
  // Calculate field at each (used) mesh point due to all other (used) mesh points
  //
  for (ix=0; ix<nCellsUsed; ix++) {
    for (iy=0; iy<nCellsUsed; iy++) {
      // Don't include (u,v) cell
      if (!((meshused[iy].x==meshused[ix].x) && (meshused[iy].y==meshused[ix].y))) {
        //
        // Calc components of field at (u,v) due to mass at (x,y) and add on
        //
        mesh[meshused[ix].x][meshused[ix].y].g.x +=
          mesh[meshused[iy].x][meshused[iy].y].m *
          LookupRx(meshused[iy].x-meshused[ix].x,meshused[iy].y-meshused[ix].y);
        mesh[meshused[ix].x][meshused[ix].y].g.y +=
          mesh[meshused[iy].x][meshused[iy].y].m *
          LookupRy(meshused[iy].x-meshused[ix].x,meshused[iy].y-meshused[ix].y);
      }
    }
  }

}

//
// FUNCTION: CheckBounds
// checks to see if particle is in mesh,
// if not, its velocity is reversed.
//
inline void CheckBounds(int is) {
  if ( (stars[is].r.x <= 0) && (stars[is].V.x < 0) ) {
    nEdgeCollisions++;
    stars[is].V.x *= -1;
  }
  if ( (stars[is].r.y <= 0) && (stars[is].V.y < 0) ) {
    nEdgeCollisions++;
    stars[is].V.y *= -1;
  }
  if ( (stars[is].r.x >= S) && (stars[is].V.x > 0) ) {
    nEdgeCollisions++;
    stars[is].V.x *= -1;
  }
  if ( (stars[is].r.y >= S) && (stars[is].V.y > 0) ) {
    nEdgeCollisions++;
    stars[is].V.y *= -1;
  }
}

//
// FUNCTION: OutputData
// Send data to screen and files
//
void OutputData(void)
{
  int is;

  //
  // Calculate number of revolutions of indicator stars
  //
  for (is=0; is<NIS; is++) {
    if (Ir[is].x<CM && stars[is+1].r.x>CM) nRevs[is] += 0.25;
    if (Ir[is].x>CM && stars[is+1].r.x<CM) nRevs[is] += 0.25;
    if (Ir[is].y<CM && stars[is+1].r.y>CM) nRevs[is] += 0.25;
    if (Ir[is].y>CM && stars[is+1].r.y<CM) nRevs[is] += 0.25;
    Ir[is].x = stars[is+1].r.x;
    Ir[is].y = stars[is+1].r.y;
  }

  //
  // Write data to files
  //
  if (OPINTDATA > 0.0) {
    if ((int)(t*D) % (int)OPINTDATA == 0) {
      WriteDataFile();
    }
  }

  if (OPINTBIN > 0.0) {
    if ((int)(t*D) % (int)OPINTBIN == 0) {

      int ix,iy;

      /*

      sprintf(szBuffer,"\nCONFIGURATION at Dt,%e\n",(double)D*t);
      _write(hGraphFile,szBuffer,lstrlen(szBuffer));

      //
      // Write density graph
      //
      sprintf(szBuffer,"\nMASS DENSITY DISTRIBUTION\nr,density\n");
      _write(hGraphFile,szBuffer,lstrlen(szBuffer));
      for (is=1; is<200; is++) {
        sprintf(szBuffer,"%.2e,%.2e\n",(double)S*is/400.0,dbins[is]*400.0/(is*d*S));
        _write(hGraphFile,szBuffer,lstrlen(szBuffer));
      }

      //
      // Write stars data
      //
      sprintf(szBuffer,"\nSTARS\ntype,mass,r,V,Ek,rx,ry\n");
      _write(hGraphFile,szBuffer,lstrlen(szBuffer));
      for (is=0; is<1000; is++) {
        sprintf(szBuffer,"%d,%.2e,%.4e,%.4e,%.4e,%.2e,%.2e\n",
        stars[is].c,stars[is].m,
        radius(stars[is].r.x-CM,stars[is].r.y-CM),
        radius(stars[is].V.x,stars[is].V.y),
        CalcEnergy(is),
        stars[is].r.x-CM,stars[is].r.y-CM);
        _write(hGraphFile,szBuffer,lstrlen(szBuffer));
      }

      //
      // Write field at mesh points
      //
      sprintf(szBuffer,"\nMESH FIELD\n");
      _write(hGraphFile,szBuffer,lstrlen(szBuffer));
      for (ix=0; ix<C; ix++) {
        for (iy=0; iy<C; iy++) {
          double F = sqrt((mesh[ix][iy].g.x*mesh[ix][iy].g.x)+
            (mesh[ix][iy].g.y*mesh[ix][iy].g.y));
          sprintf(szBuffer,"%.3e,",F);
          _write(hGraphFile,szBuffer,lstrlen(szBuffer));
        }
        sprintf(szBuffer,"\n");
        _write(hGraphFile,szBuffer,lstrlen(szBuffer));
      }

      //
      // Write mass in mesh cells
      //
      sprintf(szBuffer,"\nMESH MASS\n");
      _write(hGraphFile,szBuffer,lstrlen(szBuffer));
      for (ix=0; ix<C; ix++) {
        for (iy=0; iy<C; iy++) {
          sprintf(szBuffer,"%.2e,",mesh[ix][iy].m);
          _write(hGraphFile,szBuffer,lstrlen(szBuffer));
        }
        sprintf(szBuffer,"\n");
        _write(hGraphFile,szBuffer,lstrlen(szBuffer));
      }
      sprintf(szBuffer,"END\n");
      _write(hGraphFile,szBuffer,lstrlen(szBuffer));

    */

      //
      // Write to bitmap
      //
      #define BMPDIM 256
      RGBTRIPLE bits[BMPDIM][BMPDIM];
      BITMAPINFOHEADER bmih;
      BITMAPFILEHEADER bmfh;
      bmfh.bfType = 0x4D42;
      bmfh.bfReserved1=0;
      bmfh.bfReserved2=0;
      bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
      bmfh.bfSize = bmfh.bfOffBits + (3*BMPDIM*BMPDIM);
      bmih.biSize = sizeof(BITMAPINFOHEADER);
      bmih.biWidth = BMPDIM;
      bmih.biHeight = BMPDIM;
      bmih.biPlanes = 1;
      bmih.biBitCount = 24;
      bmih.biCompression = BI_RGB;
      bmih.biSizeImage = 0;
      bmih.biXPelsPerMeter = 300;
      bmih.biYPelsPerMeter = 300;
      bmih.biClrUsed =0;
      bmih.biClrImportant =0;
      for (ix=0; ix<BMPDIM; ix++) {
        for (iy=0; iy<BMPDIM; iy++) {
          bits[ix][iy].rgbtBlue =255;
          bits[ix][iy].rgbtGreen =255;
          bits[ix][iy].rgbtRed =255;
        }
      }
      for (is=N; is>=0; is--) {
        int y = (int)(stars[is].r.x*BMPDIM/S);
        int x = (int)(BMPDIM-(stars[is].r.y*BMPDIM/S));
        if ((x<BMPDIM && y<BMPDIM) && (x>=0 && y>=0)) {
          if (is == 2) {
            int ii;
            for (ii=-5; ii<=5; ii++) {
              bits[x+ii][y].rgbtBlue = 255;
              bits[x+ii][y].rgbtGreen = 0;
              bits[x+ii][y].rgbtRed = 0;
            }
            for (ii=-5; ii<=5; ii++) {
              bits[x][y+ii].rgbtBlue = 255;
              bits[x][y+ii].rgbtGreen = 0;
              bits[x][y+ii].rgbtRed = 0;
            }
          } else {
            int cnew = bits[x][y].rgbtBlue;
            cnew -= 5;
            if (cnew < 0) {
        cnew=0;
      }
            bits[x][y].rgbtBlue += cnew;
            bits[x][y].rgbtGreen += cnew;
            bits[x][y].rgbtRed += cnew;
          }
        }
    }
      int hBMPFile;
      char szFileName[MAX_PATH];
      sprintf(szFileName,"%03u.bmp",(int)(t*D/1e6));
      hBMPFile= _open(szFileName,_O_RDWR | _O_BINARY | _O_CREAT | _O_TRUNC,_S_IREAD | _S_IWRITE);
      _write(hBMPFile,&bmfh,sizeof(BITMAPFILEHEADER));
      _write(hBMPFile,&bmih,sizeof(BITMAPINFOHEADER));
      _write(hBMPFile,&bits,3*BMPDIM*BMPDIM);
      _close(hBMPFile);
    }
  }
}

//
// FUNCTION: WriteDataFile
// Writes entry in temporal data file
//
void WriteDataFile(void)
{
  int is;

  //
  // Write file header if t=0
  //
  if (t==0) {
    sprintf(
      szBuffer,
      "GALAXY CONFIGURATION DATA FILE\nProduced by Galaxy Simulation\nWritten by Andrew Wedgbury\naw@wedger.demon.co.uk\n\n"
    );
    _write(hDataFile,szBuffer,lstrlen(szBuffer));

    sprintf(
      szBuffer,
      "N,%d\nD,%e\nR0,%e\nRc,%e\nMGX,%e\nM0,%e\nMh,%e\nVDISP,%e\nC,%d\nS,%e\nd,%e\n",
      (int)N,
      (double)D,
      (double)R0,
      (double)Rc,
      (double)MGX,
      (double)M0,
      (double)Mh,
      (double)VDISP,
      (int)C,
      (double)S,
      (double)d
    );
    _write(hDataFile,szBuffer,lstrlen(szBuffer));

    sprintf(
      szBuffer,
      "Dt,t,Hits,Mesh,Ek,Ep,Etot,R50,R90,R100\n",
      (int)N,
      (double)D,
      (double)R0,
      (double)Rc,
      (double)MGX,
      (double)M0,
      (double)Mh,
      (double)VDISP,
      (int)C,
      (double)S,
      (double)d
    );
    _write(hDataFile,szBuffer,lstrlen(szBuffer));
  }

  //
  // Get 50% 90% and 100% mass radii
  //
  double Utot=0.0,U=0.0,R50,R90,R100;
  for (is=1; is<200; is++) {
    Utot += dbins[is]*400.0/(is*d*S);
  }
  for (is=1; is<200; is++) {
    U += dbins[is]*400.0/(is*d*S);
    if (U > Utot/2.0) {
      R50 = (double)S*is/400.0;
      break;
    }
  }
  for (; is<200; is++) {
    U += dbins[is]*400.0/(is*d*S);
    if (U > 1.0-(Utot/9.0)) {
      R90 = (double)S*is/400.0;
      break;
    }
  }
  for (; is<200; is++) {
    U += dbins[is]*400.0/(is*d*S);
    if (U >= Utot) {
      R100 = (double)S*is/400.0;
      break;
    }
  }

  //
  // Write row for this t
  //
  sprintf(
    szBuffer,
    "%.2e,%d,%d,%d,%.6e,%.6e,%e,%.6e,%.6e,%.6e\n",
    (double)D*t,
    (int)t,
    (int)nEdgeCollisions,
    (int)(100.0*nCellsUsed/((C-1)*(C-1))),
    Ek,
    Ep,
    Ep+Ek,
    R50,
    R90,
    R100
  );
  _write(hDataFile,szBuffer,lstrlen(szBuffer));
}

//
// FUNCTION: CalcEnergy
// Calculates kinetic energy of star
//
inline double CalcEnergy(int is)
{
  return 0.5 * stars[is].m * ((stars[is].V.x*stars[is].V.x) + (stars[is].V.y*stars[is].V.y));
}

//
// FUNCTION: InitRCache
// Initializes Rx and Ry arrays by precalculating values
//
void InitRCache(void)
{
  int ix,iy;
  RxCache[0][0]= 0;
  RyCache[0][0]= 0;
  for (ix=0; ix<C; ix++) {
    for (iy=0; iy<C; iy++) {
      if ((ix+iy)!=0) {
        double x = d*ix;
        double y = d*iy;
        double r = radius(x,y);
        RxCache[ix][iy]= G*x / (r*r*r);
        RyCache[ix][iy]= G*y / (r*r*r);
      }
    }
  }
}

//
// FUNCTION: LookupRx
// Looks up cached Rx value in array
//
inline double LookupRx(int ix, int iy)
{
  if ((ix>=0) && (iy>=0)) return RxCache[ix][iy];
  if ((ix<=0) && (iy>=0)) return -RxCache[-ix][iy];
  if ((ix>=0) && (iy<=0)) return RxCache[ix][-iy];
  if ((ix<=0) && (iy<=0)) return -RxCache[-ix][-iy];
  OutputDebugString("Rx lookup failed\n");
  return 1.0;
}

//
// FUNCTION: LookupRy
// Looks up cached Ry value in array
//
inline double LookupRy(int ix, int iy)
{
  if ((ix>=0) && (iy>=0)) return RyCache[ix][iy];
  if ((ix<=0) && (iy>=0)) return RyCache[-ix][iy];
  if ((ix>=0) && (iy<=0)) return -RyCache[ix][-iy];
  if ((ix<=0) && (iy<=0)) return -RyCache[-ix][-iy];
  OutputDebugString("Ry lookup failed\n");
  return 1.0;
}

//
// FUNCTION: main
// Entry point
//
int main()
{
  lstrcpy(szDataFile,"galaxy");
  d = (double)S/C;
  CM = (double)(d*C/2.0);
  InitRCache();
  xs = WINDOWX/S;
  ys = WINDOWY/S;

  OutputData();

  while (1) {
  Timestep();
  }

  _close(hDataFile);
  _close(hBinFile);
  return 0;
}

